//                       PMCTestA.cpp                     2006-12-27 Agner Fog
//
//                     PMC Test program
//
// This program is intended for optimizing a little piece of code written in 
// C++ or assembly. The code to test is inserted at the indicated place in
// PMCTestB.cpp or PMCTestB32.asm.
//
// The code to test will be executed REPETITIONS times and the test results
// will be output for each repetition. This program measures how many clock
// cycles the code to test takes in each repetition. Furthermore, it is 
// possible to set a number of Performance Monitor Counters (PMC) to count 
// the number of micro-operations (uops), cache misses, branch mispredictions,
// etc.
// 
// The setup of the Performance Monitor Counters is microprocessor-specific.
// The specifications for PMC setup for each microprocessor family is defined
// in the tables SCounterDefinition and CounterTypesDesired in file 
// PMCTestA.cpp and PMCTestB**.
//
// Compile and link PMCTestA.cpp and PMCTestB.cpp or PMCTestB32.asm for console
// mode Windows, 32-bits. The driver files MSRDriver32.sys and MSRDriver64.sys 
// must be available in the same directory. These drivers are needed for access 
// to setup the PMCs.
// You must have sufficient user privileges to load a driver.
//
//  2005-2006 GNU General Public License www.gnu.org/copyleft/gpl.html
//////////////////////////////////////////////////////////////////////////////

//#include <tchar.h>
#include "MSRDriver.h"
#include "PMCTest.h"


// main
int main(int argc, char* argv[])
{
   int repetitions, repi, i;

   // create CCounters instance and interpret CounterTypesDesired list 
	CCounters MSRCounters;

   // print heading
   printf("\n       Clock  ");
   for (i=0; i < NumCounters; i++) printf("%10s ", MSRCounters.CounterNames[i]);

   // set high priority to avoid interrupts during test
	HANDLE hProcess = GetCurrentProcess();
	SetPriorityClass(hProcess, REALTIME_PRIORITY_CLASS);

   // start MSR counters
   MSRCounters.StartCounters();

   // run the test code
   repetitions = TestLoop();

   // stop MSR counters
   MSRCounters.StopCounters();
	
   // set priority back normal
	SetPriorityClass(hProcess, NORMAL_PRIORITY_CLASS);

	// Print results
   for (repi = 0; repi < repetitions; repi++) {
      printf("\n%12i ", ClockResults[repi]);
      for (i=0; i < NumCounters; i++) {         
         printf("%10i ", PMCResults[repi][i]);
      }
   }

   printf("\n");

	return 0;
}


//////////////////////////////////////////////////////////////////////
//        CMSRDriver class member functions
//////////////////////////////////////////////////////////////////////

// constructor
CMSRDriver::CMSRDriver() {
   
   // Define Driver filename
   if (Need64BitDriver()) {
      DriverFileName = "MSRDriver64";
   }
   else {
      DriverFileName = "MSRDriver32";
   }
   
   // Define driver symbolic link name
   DriverSymbolicName = "\\\\.\\slMSRDriver";

    // initialize
   scm = NULL;                        
   service = NULL;
   hDriver = NULL;
}

// destructor
CMSRDriver::~CMSRDriver() {
   // unload driver if not already unloaded
   if (scm) UnloadDriver();
}


// load driver
int CMSRDriver::LoadDriver() {
   char DriverFileNameE[MAX_PATH], DriverFilePath[MAX_PATH], *p;

   // Open connection to Windows Service Control Manager (SCM)
   scm = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
   if(scm != NULL) {
      
      // Get the full path of the driver file name
      strcpy(DriverFileNameE, DriverFileName);
      strcat(DriverFileNameE, ".sys");           // append .sys to DriverName
      ::GetFullPathName(DriverFileNameE, MAX_PATH, DriverFilePath, &p);
      
      // Load the driver into memory and call its entry rutine
      service = ::CreateService(scm, DriverFileNameE, "MSR driver",
         SERVICE_START + SERVICE_STOP + DELETE, SERVICE_KERNEL_DRIVER,
         SERVICE_DEMAND_START, SERVICE_ERROR_IGNORE, DriverFilePath, 
         NULL, NULL, NULL, NULL, NULL);
      
      if(service == NULL) {
         // cannot load driver. 
         // Unload it in case it was not unloaded last time
         service = ::OpenService(scm, DriverFileNameE, SERVICE_ALL_ACCESS);
         SERVICE_STATUS ss;
         ::ControlService(service, SERVICE_CONTROL_STOP, &ss);         
         ::DeleteService(service);
         ::CloseServiceHandle(service);

         // try again to load driver
         service = ::CreateService(scm, DriverFileNameE, "MSR driver",
            SERVICE_START + SERVICE_STOP + DELETE, SERVICE_KERNEL_DRIVER,
            SERVICE_DEMAND_START, SERVICE_ERROR_IGNORE, DriverFilePath, 
            NULL, NULL, NULL, NULL, NULL);

         if(service == NULL) {
            // still cannot load driver
            ::CloseServiceHandle(scm); scm = 0;
            return -3;
         }
      }
      
      // Start the service
      if(!::StartService(service, 0, NULL)) {
         // error
         int ErrNo = ::GetLastError();
         // stop service and cleanup
         service = ::OpenService(scm, DriverFileName, SERVICE_ALL_ACCESS);
         SERVICE_STATUS ss;
         ::ControlService(service, SERVICE_CONTROL_STOP, &ss);         
         ::DeleteService(service);
         ::CloseServiceHandle(service);         
         ::CloseServiceHandle(scm); scm = 0;
         return ErrNo;
      }
      
      if(service) {

         // get handle to driver
         hDriver = ::CreateFile(DriverSymbolicName, GENERIC_READ + GENERIC_WRITE,
            0, NULL, OPEN_EXISTING, 0, NULL);
         
         if(hDriver == NULL || hDriver == INVALID_HANDLE_VALUE) return -5;
      }
      else return -2;
   }
   else return -1;
   
   return 0;
}


// unload driver
int CMSRDriver::UnloadDriver() {
   if(scm == NULL || service == NULL || hDriver == NULL) {
      return -6;}
   
   ::CloseHandle(hDriver); hDriver = NULL;
   
   SERVICE_STATUS ss;
   ::ControlService(service, SERVICE_CONTROL_STOP, &ss);
   
   ::DeleteService(service);
   ::CloseServiceHandle(service); service = NULL;   
   ::CloseServiceHandle(scm); scm = NULL;
   
   return 0;
}


// Access MSR or control registers
int CMSRDriver::AccessRegisters(void * pnIn, int nInLen, void * pnOut, int nOutLen) {
   if (nInLen <= 0) return 0;

   const int DeviceType = 0x22; // FILE_DEVICE_UNKNOWN;
   const int Function = 0x800;
   const int Method = 0; // METHOD_BUFFERED;
   const int Access = 1 | 2; // FILE_READ_ACCESS | FILE_WRITE_ACCESS;
   const int IOCTL_MSR_DRIVER = DeviceType << 16 | Access << 14 | Function << 2 | Method;

   DWORD len = 0;
   
   // This call results in a call to the driver rutine DispatchControl()
   int res = ::DeviceIoControl(hDriver, IOCTL_MSR_DRIVER, pnIn, nInLen,
      pnOut, nOutLen, &len, NULL);
   if (!res) { 
      // error
      return GetLastError();}
   return 0;
}


// Access MSR or control registers by queue
int CMSRDriver::AccessRegisters(CMSRInOutQue & q) {
   // number of bytes in/out
   int n = q.GetSize() * sizeof(SMSRInOut);  
   if (n <= 0) return 0;
   return AccessRegisters(q.queue, n, q.queue, n);
}


// get driver name
LPCTSTR CMSRDriver::GetDriverName() {
   return DriverFileName;
}

// read a MSR register
int CMSRDriver::MSRRead(int r) {
   SMSRInOut a;
   a.msr_command = MSR_READ;
   a.register_number = r;
   a.value = 0;
   AccessRegisters(&a,sizeof(a),&a,sizeof(a));
   return a.val[0];}

// write a MSR register
int CMSRDriver::MSRWrite(int r, int val) {
   SMSRInOut a;
   a.msr_command = MSR_WRITE;
   a.register_number = r;
   a.value = val;
   return AccessRegisters(&a,sizeof(a),&a,sizeof(a));}

// read a control register cr0 or cr4
size_t CMSRDriver::CRRead(int r) {
   if (r != 0 && r != 4) return -11;
   SMSRInOut a;
   a.msr_command = CR_READ;
   a.register_number = r;
   a.value = 0;
   AccessRegisters(&a,sizeof(a),&a,sizeof(a));
   return size_t(a.value);}

// write a control register cr0 or cr4
int CMSRDriver::CRWrite(int r, size_t val) {
   if (r != 0 && r != 4) return -12;
   SMSRInOut a;
   a.msr_command = CR_WRITE;
   a.register_number = r;
   a.value = val;
   return AccessRegisters(&a,sizeof(a),&a,sizeof(a));}


typedef BOOL (WINAPI *LPFN_ISWOW64PROCESS) (HANDLE, PBOOL);
int CMSRDriver::Need64BitDriver() {
   // tell whether we need 32 bit or 64 bit driver
   // return value:
   // 0: running in 32 bits Windows
   // 1: running 32 bits application in 64 bits Windows
   // 2: running in 64 bits Windows

#ifdef _WIN64
   return 2;
#else

   LPFN_ISWOW64PROCESS fnIsWow64Process = 
      (LPFN_ISWOW64PROCESS)GetProcAddress(
         GetModuleHandle("kernel32"),"IsWow64Process");
   
   if (fnIsWow64Process) {
      
      BOOL bIsWow64 = FALSE;
      
      if (!fnIsWow64Process(GetCurrentProcess(),&bIsWow64)) {
         return 0;
      }
      return bIsWow64;
   }
   return 0;
#endif
}



//////////////////////////////////////////////////////////////////////////////
//        CMSRInOutQue class member functions
//////////////////////////////////////////////////////////////////////////////

// constructor
CMSRInOutQue::CMSRInOutQue() {
   n = 0;
}

// put data record in queue
int CMSRInOutQue::put (EMSR_COMMAND msr_command, unsigned int register_number,
                       unsigned int value_lo, unsigned int value_hi) {

   if (n >= MAX_QUE_ENTRIES) return -10;

   queue[n].msr_command = msr_command;
   queue[n].register_number = register_number;
   queue[n].val[0] = value_lo;
   queue[n].val[1] = value_hi;
   n++;
   return 0;
}


//////////////////////////////////////////////////////////////////////////////
//        CCounters class member functions
//////////////////////////////////////////////////////////////////////////////

// constructor
CCounters::CCounters() {
   // get number of counter definitions in table
   int n = 0, CounterType; 
   char * err;
   NumCounters = 0;
   while (CounterDefinitions[n].ProcessorFamily || CounterDefinitions[n].CounterType) n++;
   NumCounterDefinitions = n;
   
   // get processor family
   family = GetProcessorFamily();

   // get all counter requests
   for (int i = 0; i < MaxNumCounters; i++) {
      CounterType = CounterTypesDesired[i];
      err = DefineCounter(CounterType);
      if (err) {
         printf("\nCannot make counter %i. %s\n", i+1, err);
      }
   }  

   // load driver
   int ErrNo = msr.LoadDriver();
   if(ErrNo) {
      printf("\nError %i: Cannot load driver %s.sys\n", ErrNo, msr.GetDriverName());
      NumCounters = 0; exit (1);
   }
}

// request a counter setup
// (return value is error message)
char * CCounters::DefineCounter(int CounterType) {
   if (CounterType == 0) return NULL;
   int i;
   SCounterDefinition * p;
   
   // search for matching counter definition
   for (i=0, p = CounterDefinitions; i < NumCounterDefinitions; i++, p++) {
      if ((p->ProcessorFamily & family) && p->CounterType == CounterType) {
         // match found
         break;
      }
   }
   if (i >= NumCounterDefinitions) return "No matching counter definition found"; // not found in list
   return DefineCounter(*p);
}

// request a counter setup
// (return value is error message)
char * CCounters::DefineCounter(SCounterDefinition & CDef) {
   int i, counternr, a, reg, eventreg, tag;

   if ( !(CDef.ProcessorFamily & family)) return "Counter not defined for present microprocessor family";
   if (NumCounters >= MaxNumCounters) return "Too many counters";

   if (CDef.CounterLast < CDef.CounterFirst) {
      CDef.CounterLast = CDef.CounterFirst;
   }
   
   // find vacant counter
   for (counternr = CDef.CounterFirst; counternr <= CDef.CounterLast; counternr++) {
      // check if this counter register is already in use
      for (i = 0; i < NumCounters; i++) {
         if (counternr == (Counters[i] & 0x7fffffff)) {
            // this counter is already in use, find another
            goto USED;
         }
      }
      if (family == INTEL_P4) {
         // check if the corresponding event register ESCR is already in use
         eventreg = GetP4EventSelectRegAddress(counternr, CDef.EventSelectReg); 
         for (i = 0; i < NumCounters; i++) {
            if (EventRegistersUsed[i] == eventreg) {
               goto USED;
            }
         }
      }

      // vacant counter found. stop searching
      break;

      USED:;
      // this counter is occupied. keep searching
   }

   if (counternr > CDef.CounterLast) {
      // no vacant counter found
      return "Counter registers are already in use";
   }

   // vacant counter found. Save name   
   CounterNames[NumCounters] = CDef.Description;

   //put MSR commands for this counter in queue   
   switch (family) {

   case INTEL_P1MMX:
      // Pentium 1 and Pentium MMX
      a = CDef.Event | (CDef.EventMask << 6);
      if (counternr == 1) a = EventRegistersUsed[0] | (a << 16);
      queue1.put(MSR_WRITE, 0x11, a);
      queue2.put(MSR_WRITE, 0x11, 0);
      queue1.put(MSR_WRITE, 0x12+counternr, 0);
      queue2.put(MSR_WRITE, 0x12+counternr, 0);
      EventRegistersUsed[0] = a;
      break;

   case INTEL_CORE2:
      // Intel Core 2
      if (counternr & 0x40000000) {
         // This is a fixed function counter
         static int FixedCountersEnabled = 0;
         if (!(FixedCountersEnabled++)) {
            // Enable fixed function counters
            queue1.put(MSR_WRITE, 0x38F, 0x3, 0x7);
            queue1.put(MSR_WRITE, 0x38D, 0x333); // 0x111=privileged level only, 0x222=user level only
            queue2.put(MSR_WRITE, 0x38D, 0x000);
         }
         break;
      }
      // All other counters continue in next case:

   case INTEL_P23: case INTEL_PM: case INTEL_CORE: 
      // Pentium Pro, Pentium II, Pentium III, Pentium M, Core 1, (Core 2 continued):
      a = CDef.Event | (CDef.EventMask << 8) | (1 << 16) | (1 << 22);
      if (family >= INTEL_CORE2) a |= (1 << 14);  // Means this core only
      eventreg = 0x186 + counternr;
      reg = 0xc1 + counternr;
      queue1.put(MSR_WRITE, eventreg, a);
      queue2.put(MSR_WRITE, eventreg, 0);
      queue1.put(MSR_WRITE, reg, 0);
      queue2.put(MSR_WRITE, reg, 0);
      break;

   case INTEL_P4:
      // Pentium 4 and Pentium 4 with EM64T
      // ESCR register
      eventreg = GetP4EventSelectRegAddress(counternr, CDef.EventSelectReg); 
      tag = 1;
      a = 0x1C | (tag << 5) | (CDef.EventMask << 9) | (CDef.Event << 25);
      queue1.put(MSR_WRITE, eventreg, a);
      queue2.put(MSR_WRITE, eventreg, 0);
      // remember this event register is used
      EventRegistersUsed[NumCounters] = eventreg;
      // CCCR register
      reg = counternr + 0x360;
      a = (1 << 12) | (3 << 16) | (CDef.EventSelectReg << 13);
      queue1.put(MSR_WRITE, reg, a);
      queue2.put(MSR_WRITE, reg, 0);
      // reset counter register
      reg = counternr + 0x300;
      queue1.put(MSR_WRITE, reg, 0);
      queue2.put(MSR_WRITE, reg, 0);
      // set high bit for fast RDPMC
      counternr |= 0x80000000;
      break;

   case AMD_ATHLON: case AMD_ATHLON64:
      // AMD Athlon, Athlon 64, Opteron
      a = CDef.Event | (CDef.EventMask << 8) | (1 << 16) | (1 << 22);
      eventreg = 0xc0010000 + counternr;
      reg = 0xc0010004 + counternr;
      queue1.put(MSR_WRITE, eventreg, a);
      queue2.put(MSR_WRITE, eventreg, 0);
      queue1.put(MSR_WRITE, reg, 0);
      queue2.put(MSR_WRITE, reg, 0);
      break;

   default:
      return "No counters defined for present microprocessor family";
   }

   // save counter register number in Counters list
   Counters[NumCounters++] = counternr;

   return NULL; // NULL = success
}

// start counting
void CCounters::StartCounters() {
   msr.AccessRegisters(queue1);
}

// stop and reset counters
void CCounters::StopCounters() {
   msr.AccessRegisters(queue2);
}

// translate event select register number to register address for P4 processor
int CCounters::GetP4EventSelectRegAddress(int CounterNr, int EventSelectNo) {
   // On Pentium 4 processors, the Event Select Control Registers (ESCR) are
   // identified in a very confusing way. Each ESCR has both an ESCRx-number which 
   // is part of its name, an event select number to specify in the Counter 
   // Configuration Control Register (CCCR), and a register address to specify 
   // in the WRMSR instruction.
   // This function gets the register address based on table 15-6 in Intel manual 
   // 25366815, IA-32 Intel Architecture Software Developer's Manual Volume 3: 
   // System Programming Guide, 2005.
   // Returns -1 if error.
   static int TranslationTables[4][8] = {
      {0x3B2, 0x3B4, 0x3AA, 0x3B6, 0x3AC, 0x3C8, 0x3A2, 0x3A0}, // counter 0-3
      {0x3C0, 0x3C4, 0x3C2,    -1,    -1,    -1,    -1,    -1}, // counter 4-7
      {0x3A6, 0x3A4, 0x3AE, 0x3B0,    -1, 0x3A8,    -1,    -1}, // counter 8-11
      {0x3BA, 0x3CA, 0x3BC, 0x3BE, 0x3B8, 0x3CC, 0x3E0,    -1}};// counter 12-17

   unsigned int n = CounterNr;
   if (n > 17) return -1;
   if (n > 15) n -= 3;
   if ((unsigned int)EventSelectNo > 7) return -1;

   int a = TranslationTables[n/4][EventSelectNo];
   if (a < 0) return a;
   if (n & 2) a++;
   return a;
}


//////////////////////////////////////////////////////////////////////////////
//             list of counter definitions
//////////////////////////////////////////////////////////////////////////////
// How to add new entries to this list:
//
// Warning: Be sure to save backup copies of your files before you make any 
// changes here. A wrong register number can result in a crash in the driver.
// This results in a blue screen and possibly loss of your most recently
// modified file.
//
// Set CounterType to any vacant id number. Use the same id for similar events
// in different processor families. The other fields depend on the processor 
// family as follows:
//
// Pentium 1 and Pentium MMX:
//    Set ProcessorFamily = INTEL_P1MMX.
//    CounterFirst = 0, CounterLast = 1, Event = Event number,
//    EventMask = Counter control code.
//
// Pentium Pro, Pentium II, Pentium III, Pentium M, Core Solo/Duo
//    Set ProcessorFamily = INTEL_P23M for events that are valid for all these
//    processors or INTEL_PM or INTEL_CORE for events that only apply to 
//    one processor.
//    CounterFirst = 0, CounterLast = 1, 
//    Event = Event number, EventMask = Unit mask.
//
// Core 2
//    Set ProcessorFamily = INTEL_CORE2.
//    Fixed function counters:
//    CounterFirst = 0x40000000 + MSR Address - 0x309. (Intel manual Oct. 2006 is wrong)
//    All other counters:
//    CounterFirst = 0, CounterLast = 1, 
//    Event = Event number, EventMask = Unit mask.
//
// Pentium 4 and Pentium 4 with EM64T (Netburst):
//    Set ProcessorFamily = INTEL_P4.
//    Look in Software Developer's Manual vol. 3 appendix A, table of 
//    Performance Monitoring Events.
//    Set CounterFirst and CounterLast to the range of possible counter
//    registers listed under "Counter numbers per ESCR".
//    Set EventSelectReg to the value listed for "CCCR Select".
//    Set Event to the value indicated for "ESCR Event Select".
//    Set EventMask to a combination of the relevant bits for "ESCR Event Mask".
//    You don't need the table named "Performance Counter MSRs and Associated
//    CCCR and ESCR MSRs". This table is already implemented in the function
//    CCounters::GetP4EventSelectRegAddress.
//
// AMD Athlon 64, Opteron
//    Set ProcessorFamily = AMD_ATHLON64.
//    CounterFirst = 0, CounterLast = 3, Event = Event mask,
//    EventMask = Unit mask.
//

extern "C" SCounterDefinition CounterDefinitions[] = {
//  id       cpu      countregs eventreg event  mask   name
   {100,    INTEL_P4,  4,   7,     0,      9,      7,  "Uops"     }, // uops from any source
   {101,    INTEL_P4,  4,   7,     0,      9,      2,  "UopsTC"   }, // uops from trace cache
   {102,    INTEL_P4,  4,   7,     0,      9,      1,  "UopsDec"  }, // uops directly from decoder
   {103,    INTEL_P4,  4,   7,     0,      9,      4,  "UopsMCode"}, // uops from microcode ROM
   {110,    INTEL_P4, 12,  17,     4,      1,      1,  "UopsNB"   }, // uops non-bogus
   {111,    INTEL_P4, 12,  17,     4,      2,   0x0c,  "UopsBogus"}, // uops bogus

   {150,    INTEL_P4,  8,  11,     1,      4, 0x8000,  "UopsFP"   }, // uops floating point, except move etc.
   {151,    INTEL_P4,  8,  11,     1,   0x2e,      8,  "UopsFPMov"}, // uops floating point and SIMD move
   {152,    INTEL_P4,  8,  11,     1,   0x2e,   0x10,  "UopsFPLd" }, // uops floating point and SIMD load
   {160,    INTEL_P4,  8,  11,     1,      2, 0x8000,  "UopsMMX"  }, // uops 64-bit MMX
   {170,    INTEL_P4,  8,  11,     1,   0x1a, 0x8000,  "UopsXMM"  }, // uops 128-bit integer XMM

   {200,    INTEL_P4, 12,  17,     5,      6,   0x0f,  "Branch"   }, // branches
   {201,    INTEL_P4, 12,  17,     5,      6,   0x0c,  "BrTaken"  }, // branches taken
   {202,    INTEL_P4, 12,  17,     5,      6,   0x03,  "BrNTaken" }, // branches not taken
   {203,    INTEL_P4, 12,  17,     5,      6,   0x05,  "BrPredict"}, // branches predicted
   {204,    INTEL_P4, 12,  17,     4,      3,   0x01,  "BrMispred"}, // branches mispredicted
   {210,    INTEL_P4,  4,   7,     2,      5,   0x02,  "CondJMisp"}, // conditional jumps mispredicted
   {211,    INTEL_P4,  4,   7,     2,      5,   0x04,  "CallMisp" }, // indirect call mispredicted
   {212,    INTEL_P4,  4,   7,     2,      5,   0x08,  "RetMisp"  }, // return mispredicted
   {220,    INTEL_P4,  4,   7,     2,      5,   0x10,  "IndirMisp"}, // indirect calls, jumps and returns mispredicted

   {310,    INTEL_P4,  0,   3,     0,      3,   0x01,  "TCMiss"   }, // trace cache miss
   {320,    INTEL_P4,  0,   3,     7,   0x0c,  0x100,  "Cach2Miss"}, // level 2 cache miss
   {321,    INTEL_P4,  0,   3,     7,   0x0c,  0x200,  "Cach3Miss"}, // level 3 cache miss
   {330,    INTEL_P4,  0,   3,     3,   0x18,   0x02,  "ITLBMiss" }, // instructions TLB Miss
   {340,    INTEL_P4,  0,   3,     2,      3,   0x3a,  "LdReplay" }, // memory load replay


//  id       cpu      countregs eventreg event  mask   name
   {  9, INTEL_P1MMX,  0,   1,     0,   0x16,      2,  "Instruct" }, // instructions executed
   { 11, INTEL_P1MMX,  0,   1,     0,   0x17,      2,  "InstVpipe"}, // instructions executed in V-pipe
   {202, INTEL_P1MMX,  0,   1,     0,   0x15,      2,  "Flush"    }, // pipeline flush due to branch misprediction or serializing event   
   {310, INTEL_P1MMX,  0,   1,     0,   0x0e,      2,  "CodeMiss" }, // code cache miss
   {311, INTEL_P1MMX,  0,   1,     0,   0x29,      2,  "DataMiss" }, // data cache miss
   

//  id       cpu      countregs eventreg event  mask   name
   {  9, INTEL_P23M,   0,   1,     0,   0xc0,      0,  "Instruct" }, // instructions executed
   { 10, INTEL_P23M,   0,   1,     0,   0xd0,      0,  "IDecode"  }, // instructions decoded
   { 20, INTEL_P23M,   0,   1,     0,   0x80,      0,  "IFetch"   }, // instruction fetches
   { 21, INTEL_P23M,   0,   1,     0,   0x86,      0,  "IFetchStl"}, // instruction fetch stall
   { 22, INTEL_P23M,   0,   1,     0,   0x87,      0,  "ILenStal" }, // instruction length decoder stalls
   {100, INTEL_P23M,   0,   1,     0,   0xc2,      0,  "Uops"     }, // microoperations
   {100, INTEL_PM,     0,   1,     0,   0xc2,      0,  "Uops(F)"  }, // microoperations in fused domain
   {110, INTEL_PM,     0,   1,     0,   0xa0,      0,  "Uops(UF)" }, // unfused microoperations submitted to execution units (Undocumented counter!)
   {104, INTEL_PM,     0,   1,     0,   0xda,      0,  "UopsFused"}, // fused uops
   {115, INTEL_PM,     0,   1,     0,   0xd3,      0,  "SynchUops"}, // stack synchronization uops
   {121, INTEL_P23M,   0,   1,     0,   0xd2,      0,  "PartRStl" }, // partial register access stall
   {130, INTEL_P23M,   0,   1,     0,   0xa2,      0,  "Rs Stall" }, // all resource stalls
   {201, INTEL_P23M,   0,   1,     0,   0xc9,      0,  "BrTaken"  }, // branches taken
   {204, INTEL_P23M,   0,   1,     0,   0xc5,      0,  "BrMispred"}, // mispredicted branches
   {209, INTEL_P23M,   0,   1,     0,   0xe6,      0,  "BTBMiss"  }, // static branch prediction made
   {310, INTEL_P23M,   0,   1,     0,   0x28,   0x0f,  "CodeMiss" }, // level 2 cache code fetch
   {311, INTEL_P23,    0,   1,     0,   0x29,   0x0f,  "DataMiss" }, // level 2 cache data fetch


// Core 2:
// The first three counters are fixed-function counters having their own register,
// The rest of the counters are competing for the same two registers.
//  id       cpu      countregs eventreg event  mask   name
   {1,   INTEL_CORE2,   0x40000001,  0,0,   0,      0,   "Core cyc"   }, // core clock cycles
   {2,   INTEL_CORE2,   0x40000002,  0,0,   0,      0,   "Ref cyc"    }, // Reference clock cycles
   {9,   INTEL_CORE2,   0x40000000,  0,0,   0,      0,   "Instruct"   }, // Instructions (reference counter)
   {10,  INTEL_CORE2,   0,   1,     0,   0xc0,     0x0f, "Instrust"   }, // Instructions
   {11,  INTEL_CORE2,   0,   1,     0,   0xc0,     0x01, "Read inst"  }, // Instructions involving read, fused count as one
   {12,  INTEL_CORE2,   0,   1,     0,   0xc0,     0x02, "Write ins"  }, // Instructions involving write, fused count as one
   {13,  INTEL_CORE2,   0,   1,     0,   0xc0,     0x04, "NonMem in"  }, // Instructions without memory
   {20,  INTEL_CORE2,   0,   1,     0,   0x80,      0,   "Insfetch"   }, // instruction fetches. < instructions ?
   {21,  INTEL_CORE2,   0,   1,     0,   0x86,      0,   "IFetchStl"  }, // instruction fetch stall
   {22,  INTEL_CORE2,   0,   1,     0,   0x87,      0,   "ILenStal"   }, // instruction length decoder stalls (length changing prefix)
   {23,  INTEL_CORE2,   0,   1,     0,   0x83,      0,   "IQue ful"   }, // instruction queue full
   {100, INTEL_CORE2,   0,   1,     0,   0xc2,     0x0f, "Uops"       }, // uops retired, fused domain
   {101, INTEL_CORE2,   0,   1,     0,   0xc2,     0x01, "Fused Rd"   }, // fused read uops
   {102, INTEL_CORE2,   0,   1,     0,   0xc2,     0x02, "Fused Wrt"  }, // fused write uops
   {103, INTEL_CORE2,   0,   1,     0,   0xc2,     0x04, "Macrofus"   }, // macrofused uops
   {104, INTEL_CORE2,   0,   1,     0,   0xc2,     0x07, "FusedUop"   }, // fused uops, all kinds
   {105, INTEL_CORE2,   0,   1,     0,   0xc2,     0x08, "NotFusUop"  }, // uops, not fused
   {110, INTEL_CORE2,   0,   1,     0,   0xa0,        0, "Uops UFD"   }, // uops dispatched, unfused domain. Imprecise
   {115, INTEL_CORE2,   0,   1,     0,   0xab,     0x01, "SP synch"   }, // Stack synchronization uops
   {116, INTEL_CORE2,   0,   1,     0,   0xab,     0x02, "SP engine"  }, // Stack engine additions
   {121, INTEL_CORE2,   0,   1,     0,   0xd2,     0x02, "Part.reg"   }, // Partial register synchronization, clock cycles
   {122, INTEL_CORE2,   0,   1,     0,   0xd2,     0x04, "part.flag"  }, // partial flags stall, clock cycles
   {123, INTEL_CORE2,   0,   1,     0,   0xd2,     0x08, "FP SW stl"  }, // floating point status word stall
   {130, INTEL_CORE2,   0,   1,     0,   0xd2,     0x01, "R Rd stal"  }, // ROB register read stall
   {140, INTEL_CORE2,   0,   1,     0,   0x19,     0x00, "I2FP pass"  }, // bypass delay to FP unit from int unit
   {141, INTEL_CORE2,   0,   1,     0,   0x19,     0x01, "FP2I pass"  }, // bypass delay to SIMD/int unit from fp unit (These counters cannot be used simultaneously)
   {142, INTEL_CORE2,   0,   1,     0,   0x19,     0x02, "Load pass"  }, // bypass delay to load unit. Doesn't count?
   {201, INTEL_CORE2,   0,   1,     0,   0xc4,     0x0c, "BrTaken"    }, // branches taken. (Mask: 1=pred.not taken, 2=mispred not taken, 4=pred.taken, 8=mispred taken)
   {204, INTEL_CORE2,   0,   1,     0,   0xc4,     0x0a, "BrMispred"  }, // mispredicted branches
   {205, INTEL_CORE2,   0,   1,     0,   0xe6,      0,   "BTBMiss"    }, // static branch prediction made
   {210, INTEL_CORE2,   0,   1,     0,   0x97,      0,   "BranchBu1"  }, // branch taken bubble 1
   {211, INTEL_CORE2,   0,   1,     0,   0x98,      0,   "BranchBu2"  }, // branch taken bubble 2 (these two values must be added)
   {310, INTEL_CORE2,   0,   1,     0,   0x28,     0x0f, "CodeMiss"   }, // level 2 cache code fetch
   {311, INTEL_CORE2,   0,   1,     0,   0x29,     0x0f, "DataMiss"   }, // level 2 cache data fetch
   {320, INTEL_CORE2,   0,   1,     0,   0x24,     0x00, "L2 Miss"    }, // level 2 cache miss


//  id       cpu      countregs eventreg event  mask   name
   {  9, AMD,          0,   3,     0,   0xc0,      0,  "Instruct" }, // x86 instructions executed
   {100, AMD,          0,   3,     0,   0xc1,      0,  "Uops"     }, // microoperations
   {150, AMD,          0,   3,     0,   0x00,   0x3f,  "UopsFP"   }, // microoperations in FP pipe
   {151, AMD,          0,   3,     0,   0x00,   0x09,  "FPADD"    }, // microoperations in FP ADD unit
   {152, AMD,          0,   3,     0,   0x00,   0x12,  "FPMUL"    }, // microoperations in FP MUL unit
   {153, AMD,          0,   3,     0,   0x00,   0x24,  "FPMISC"   }, // microoperations in FP Store unit
   {204, AMD,          0,   3,     0,   0xc3,      0,  "BrMispred"}, // mispredicted branches
   {310, AMD,          0,   3,     0,   0x81,      0,  "CodeMiss" }, // instruction cache misses
   {311, AMD,          0,   3,     0,   0x41,      0,  "DataMiss" }, // data cache misses

//  end of list   
   {0,      UNKNOWN,   0,   0,     0,      0,      0,    0        }  // list must end with a record of all 0
};
