comment & ---------------------------------------------------------------------
                           PMCTestB32.asm                 2006-08-14 Agner Fog
  
                          PMC Test program
                               
This program is intended for optimizing a little piece of code written in 
C++ or assembly. The code to test is inserted at the place marked
"Test code start" in PMCTestB32.asm.
All sections that can be modified by the user are marked with ###########.
  
The code to test will be executed REPETITIONS times and the test results
will be output for each repetition. This program measures how many clock
cycles the code to test takes in each repetition. Furthermore, it is 
possible to set a number of Performance Monitor Counters (PMC) to count 
the number of micro-operations (uops), cache misses, branch mispredictions,
etc.
 
The setup of the Performance Monitor Counters is microprocessor-specific.
The specifications for PMC setup for each microprocessor family is defined
in the tables SCounterDefinition and CounterTypesDesired.

Compile PMCTestA.cpp, assemble PMCTestB32.asm, and link for console mode
Windows, 32-bits. The driver files MSRDriver32.sys and MSRDriver64.sys
must be available in the same directory. This driver is needed for access 
to setup the PMCs. You must have sufficient user privileges to load the driver.

If you are running this on old processors (Pentium 1 or Pentium MMX) then
you must replace all cmov instructions with branches.

 2005-2006 GNU General Public License www.gnu.org/copyleft/gpl.html

----------------------------------------------------------------------------- &
INCLUDE UserCode.asm
; Define platform
.686
.xmm
.model flat

INCLUDELIB advapi32.lib
INCLUDELIB kernel32.lib
INCLUDELIB user32.lib
INCLUDELIB masm32.lib
;##############################################################################
;#
;#                 list of desired counter types
;#
;##############################################################################
.data
 
; Here you can select which performance monitor counters you want for your test.
; Select id numbers from the table CounterDefinitions[] in PMCTestA.cpp.

_CounterTypesDesired label DWORD
public _CounterTypesDesired
   DD     1         ; number of instructions
   DD     2         ; number of uops
   DD   201         ; code cache misses
   DD   202         ; data cache misses
    
; number of counters defined
NUMCOUNTERS = ($ - _CounterTypesDesired) / 4


;##############################################################################
;#
;#                 Define number of repetitions
;#
;##############################################################################

; Number of repetitions of test.
REPETITIONS = 20

; Maximum number of PMC counters
MAXCOUNTERS = 4            ; must match value in PMCTest.h

; Define array sizes
MAXREPEAT = REPETITIONS

;------------------------------------------------------------------------------
;
;                  global data
;
;------------------------------------------------------------------------------
.data

public _NumCounters, _MaxNumCounters, _Counters, _EventRegistersUsed
public _ClockResults, _PMCResults

; save esp for security
SaveEsp          DD    0

; Number of performance counters used
;extern "C" int NumCounters = 0;
_NumCounters     DD    0
_MaxNumCounters  DD    NUMCOUNTERS   ; tell PMCTestA.cpp length of _CounterTypesDesired

; Counter register numbers used
;extern "C" int Counters[MAXCOUNTERS];
_Counters        DD    MAXCOUNTERS dup (0)
_EventRegistersUsed DD MAXCOUNTERS dup (0)

; Results tables
ClockTemp        DD    0
PMCTemp          DD    MAXCOUNTERS dup (0)
_ClockResults    DD    MAXREPEAT dup (0)
_PMCResults      DD    (MAXREPEAT * MAXCOUNTERS) dup (0)
ClockOverhead    DD    -1
PMCOverhead      DD    MAXCOUNTERS dup (-1)
repi             DD    0               ; repetition loop counter
PMCResults_repi  DD    0               ; &PMCResults[repi][0]


;##############################################################################
;#
;#                 User data
;#
;##############################################################################

; Put any data definitions your test code needs here

UserdataSection

;------------------------------------------------------------------------------
;
;                  Test Loop
;
;------------------------------------------------------------------------------
.code

;extern "C" int TestLoop () {
; This function runs the code to test REPETITIONS times
; and reads the counters before and after each run:

IF REPETITIONS GT 4  
   repi1 = 3            ; exclude first few runs from minimum search
ELSE
   repi1 = 0
ENDIF

SERIALIZE MACRO             ; serialize CPU
       xor     eax, eax
       cpuid
ENDM

_TestLoop PROC NEAR
        push    ebx
        push    esi
        push    edi
        push    ebp
        mov     SaveEsp, esp

;##############################################################################
;#
;#                 User Initializations 
;#
;##############################################################################
; You may add any initializations your test code needs here
; esi, edi and ebp will be unchanged from here to the first Test code start

        finit                          ; reset floating point unit
        
if 0    ; change to 1 if xmm supported and you want these registers cleared
        pxor     xmm0, xmm0            ; reset all xmm registers
        pxor     xmm1, xmm1
        pxor     xmm2, xmm2
        pxor     xmm3, xmm3
        pxor     xmm4, xmm4
        pxor     xmm5, xmm5
        pxor     xmm6, xmm6
        pxor     xmm7, xmm7
endif

Initialization
;##############################################################################
;#
;#                 End of user Initializations 
;#
;##############################################################################

        
; First test loop
; Measure overhead = the test count produced by the test program itself

        mov     repi, 0
        mov     PMCResults_repi, offset ds:_PMCResults
TEST_LOOP_1:

        SERIALIZE
      
        ; Read counters
        I = 0
REPT    NUMCOUNTERS        
        mov     ecx, _Counters[I*4]
        rdpmc
        mov     PMCTemp[I*4], eax
        I = I + 1
ENDM       

        SERIALIZE

        ; read time stamp counter
        rdtsc
        mov     ClockTemp, eax

        SERIALIZE

        ; No test code here

        SERIALIZE

        ; read time stamp counter
        rdtsc
        mov     ecx, repi
        mov     _ClockResults[ecx*4], eax

        SERIALIZE

        ; Read counters
        mov     ebx, PMCResults_repi
        I = 0
REPT    NUMCOUNTERS     
        mov     ecx, _Counters[I*4]
        rdpmc
        mov     [ebx][I*4], eax
        I = I + 1
ENDM       

        SERIALIZE

        ; subtract counts before from counts after
        mov     ecx, repi
        mov     ebx, PMCResults_repi
        mov     eax, ClockTemp
        sub     _ClockResults[ecx*4], eax
        I = 0
REPT    NUMCOUNTERS
        mov     eax, PMCTemp[I*4]
        sub     [ebx][I*4], eax
        I = I + 1
ENDM       

        ; end fist test loop
        add     ecx, 1
        mov     repi, ecx
        add     ebx, MAXCOUNTERS * 4
        mov     PMCResults_repi, ebx
        cmp     ecx, REPETITIONS
        jb      TEST_LOOP_1
        
; Loop to find minimum counts:
        ; for (repi = repi1; repi < REPETITIONS; repi++) {
        mov     ecx, repi1                  ; repi = repi1
        mov     ebx, MAXCOUNTERS * 4
        imul    ebx, ecx
        add     ebx, offset ds:_PMCResults  ; &PMCResults[repi][0]
        
FIND_MINIMUM_LOOP:
        ; minimum clock overhead
        mov     eax, ClockOverhead
        mov     edx, _ClockResults[ecx*4]
        cmp     eax, edx
        cmova   eax, edx
        mov     ClockOverhead, eax

        ; minimum PMC counts
        ; if (PMCResults[repi][i] < PMCOverhead[i]) PMCOverhead[i] = PMCResults[repi][i];
        I = 0
REPT    NUMCOUNTERS
        mov     eax, PMCOverhead[I*4]
        mov     edx, [ebx][I*4]
        cmp     eax, edx
        cmova   eax, edx
        mov     PMCOverhead[I*4], eax
        I = I + 1
ENDM       
        
        ; end find minimum loop
        add     ecx, 1
        add     ebx, MAXCOUNTERS * 4
        cmp     ecx, REPETITIONS
        jb      FIND_MINIMUM_LOOP

        
; Second test loop. Includes code to test
; This must be identical to first test loop, except for the test code

        mov     repi, 0
        mov     PMCResults_repi, offset ds:_PMCResults
TEST_LOOP_2:

        SERIALIZE
      
        ; Read counters
        I = 0
REPT    NUMCOUNTERS
        mov     ecx, _Counters[I*4]
        rdpmc
        mov     PMCTemp[I*4], eax
        I = I + 1
ENDM       

        SERIALIZE

        ; read time stamp counter
        rdtsc
        mov     ClockTemp, eax

        SERIALIZE

;##############################################################################
;#
;#                 Test code start
;#
;##############################################################################

; Put the assembly code to test here or call a function in a separate module.
; 
        TestedPieceOfCode
;##############################################################################
;#
;#                 Test code end
;#
;##############################################################################

        SERIALIZE

        ; read time stamp counter
        rdtsc
        mov     ecx, repi
        mov     _ClockResults[ecx*4], eax

        SERIALIZE

        ; Read counters
        mov     ebx, PMCResults_repi
        I = 0
REPT    NUMCOUNTERS
        mov     ecx, _Counters[I*4]
        rdpmc
        mov     [ebx][I*4], eax
        I = I + 1
ENDM       

        SERIALIZE

        ; subtract counts before from counts after
        mov     ecx, repi
        mov     ebx, PMCResults_repi
        mov     eax, ClockTemp
        sub     _ClockResults[ecx*4], eax
        I = 0
REPT    NUMCOUNTERS
        mov     eax, PMCTemp[I*4]
        sub     [ebx][I*4], eax
        I = I + 1
ENDM       

        ; end second test loop
        add     ecx, 1
        mov     repi, ecx
        add     ebx, MAXCOUNTERS * 4
        mov     PMCResults_repi, ebx
        cmp     ecx, REPETITIONS
        jb      TEST_LOOP_2

; Loop to subtract overhead
        ; for (repi = 0; repi < REPETITIONS; repi++) {
        mov     ecx, 0                      ; repi = 0
        mov     ebx, offset ds:_PMCResults  ; &PMCResults[0][0]
        
SUBTRACT_OVERHEAD_LOOP:

        ; ClockResults[repi]  -= ClockOverhead;
        mov     eax, ClockOverhead
        sub     _ClockResults[ecx*4], eax

        ; PMCResults[repi][i] -= PMCOverhead[i];
        I = 0
REPT    NUMCOUNTERS
        mov     eax, PMCOverhead[I*4]
        sub     [ebx][I*4], eax
        I = I + 1
ENDM         
        ; end subtract overhead loop
        add     ecx, 1
        add     ebx, MAXCOUNTERS * 4
        cmp     ecx, REPETITIONS
        jb      SUBTRACT_OVERHEAD_LOOP

        ; clean up
        finit
        cld
        mov     esp, SaveEsp

        ; return REPETITIONS;
        mov     eax, REPETITIONS
        pop     ebp
        pop     edi
        pop     esi
        pop     ebx
        
        pushad
        
        CleanUp_End
        popad
        ret
        
_TestLoop ENDP


;------------------------------------------------------------------------------
;
;                  GetProcessorFamily function
;
;------------------------------------------------------------------------------
comment &
Returns one of the following values:

enum EProcFamily {
   UNKNOWN = 0,              // unknown. can't do performance monitoring
   INTEL_P1MMX = 1,          // Intel Pentium 1 or Pentium MMX
   INTEL_P23 = 2,            // Pentium Pro, Pentium 2, Pentium 3
   INTEL_PM = 4,             // Pentium M
   INTEL_P4 = 8,             // Pentium 4 and Pentium 4 with EM64T
   INTEL_CORE  = 0x10,       // Intel Core Solo/Duo
   INTEL_CORE2 = 0x20,       // Intel Core 2
   AMD_ATHLON = 0x1000,      // AMD Athlon
   AMD_ATHLON64 = 0x2000,    // AMD Athlon 64 or Opteron
};
&

; EProcFamily GetProcessorFamily() {
_GetProcessorFamily PROC NEAR
        push    ebx                    ; save ebx
        xor     eax, eax               ; call cpuid function 0
        cpuid                          ; get cpu information
        cmp     eax, 0
        je      RETURNVAL              ; further CPUID functions not supported. return UNKNOWN
        cmp     cl, 'n'
        je      IS_INTEL               ; Intel processor
        cmp     ch, 'A'
        je      IS_AMD                 ; AMD processor
        ; unknown processor
        xor     eax, eax               ; 0 = UNKNOWN
        jmp     RETURNVAL              ; return UNKNOWN
   
IS_INTEL:       ; Intel processor
        mov     eax, 1                 ; call cpuid function 1
        cpuid                          ; eax = family and model
        mov     ebx, eax
        and     bh,  0FH               ; BH = family
        shr     bl,  4                 ; BL = model
        xor     eax, eax               ; 0 = UNKNOWN
        cmp     bh,  5
        jb      RETURNVAL              ; family < 5. return UNKNOWN
        mov     al, 1                  ; 1 = INTEL_P1MMX
        je      RETURNVAL              ; family = 5. return INTEL_P1MMX
        mov     al, 8                  ; 8 = INTEL_P4
        cmp     bh, 0FH
        je      RETURNVAL              ; family = 15. return INTEL_P4
        mov     al, 0
        cmp     bh, 6
        jne     RETURNVAL              ; family != 6. return UNKNOWN
        mov     al, 4                  ; 4 = INTEL_PM
        cmp     bl, 9
        je      RETURNVAL              ; family = 6, model = 9. return INTEL_PM
        cmp     bl, 0DH
        je      RETURNVAL              ; family = 6, model = 13. return INTEL_PM
        mov     al, 10h                ; 10h = INTEL_CORE
        cmp     bl, 0EH
        je      RETURNVAL              ; family = 6, model = 14. return INTEL_CORE        
        mov     al, 20h                ; 20h = INTEL_CORE2
        cmp     bl, 0FH
        je      RETURNVAL              ; family = 6, model = 15. return INTEL_CORE2        
        mov     al, 2                  ; 2 = INTEL_P23
        jmp     RETURNVAL              ; family = 6, not INTEL_PM. return INTEL_P23

IS_AMD:         ; AMD processor        
        mov     eax, 1                 ; call cpuid function 1
        cpuid                          ; eax = family and model
        mov     ebx, eax
        and     bh,  0FH               ; BH = family
        shr     bl,  4                 ; BL = model
        xor     eax, eax               ; 0 = UNKNOWN
        cmp     bh, 6
        jb      RETURNVAL              ; family < 6. return UNKNOWN
        mov     ah, 10H                ; 1000H = AMD_ATHLON
        je      RETURNVAL              ; family = 6. return AMD_ATHLON
        add     eax, eax               ; 2000H = AMD_ATHLON64
        cmp     bh, 0FH
        je      RETURNVAL              ; family = 15. return AMD_ATHLON64
        xor     eax, eax               ; other. return 0 = UNKNOWN

RETURNVAL:                             ; return value is in eax
        pop     ebx
        ret
_GetProcessorFamily ENDP                

SomeOtherFunctions




END   
